#include <errno.h>

#include <csparse.h>
#include <cstats.h>
#include <string.h>
#include "tbx_timer.h"

#include "interface.h"

#include "app.h"

#ifndef PARALLEL_H
#define PARALLEL_H

#define INFO_STRLEN    20
#define COMMAND_STRLEN 80
#define PGM_STRLEN     80

#define INTRO_LETTER   1
#define COMMAND_LETTER 2
#define MATRIX_LETTER  3
#define VECTOR_LETTER  4
#define TIMING_LETTER  5

#define DO_LU          1
#define DO_SOLVE       2
#define TERMINATE      3

struct matrix_host_bin
{
   int               P;
   int               Q;
   int               p;
   int               q;

   ZIP_ADDRESSEES   *Addr_List;
   ZIP_MAILER       *Mailer;

   short             Row_Dis_Type;
   char              Row_Extra_param1_name[2];
   long              Row_Extra_param1;
   char              Row_Extra_param2_name[2];
   long              Row_Extra_param2;
   Cdistrib         *Row_Dis;
   void             *Row_Extra;
   void            (*Row_Distribution)();
   int             (*Row_Inverse_Distribution)();
   int             (*Row_Distribution_limit)();
   int             (*Row_Distribution_init)();

   short             Col_Dis_Type;
   char              Col_Extra_param1_name[2];
   long              Col_Extra_param1;
   char              Col_Extra_param2_name[2];
   long              Col_Extra_param2;
   Cdistrib         *Col_Dis;
   void             *Col_Extra;
   void            (*Col_Distribution)();
   int             (*Col_Inverse_Distribution)();
   int             (*Col_Distribution_limit)();
   int             (*Col_Distribution_init)();

   long              Pivot_Strategy;

   int               M;
   int               N;
   int               m;
   int               n;

   int             **A_element_Count;
   struct element ***A;

   int               NBUS;
   int              *Row_major_LU_element_Count;
   struct element  **Row_major_LU;

   struct element   *Pool;

};

typedef struct matrix_host_bin matrix_host_bin;

struct vector_host_bin
{
   int              P;
   int              Q;
   int              M;
   int              N;

   Cdistrib        *Row_Dis;
   void           (*Row_Distribution)();
   int            (*Row_Inverse_Distribution)();
   int            (*Row_Distribution_limit)();
   void            *Row_Extra;

   Cdistrib        *Col_Dis;
   void           (*Col_Distribution)();
   int            (*Col_Inverse_Distribution)();
   int            (*Col_Distribution_limit)();
   void            *Col_Extra;

   int             *b_element_Count;
   struct element **b;

   int               NBUS;
   struct element   *Col_major_x;
   int               x_element_Count;

   struct element   *Pool;
};

typedef struct vector_host_bin vector_host_bin;

struct element
{
   int  i;
   int  j;
   double Real;
   double Imag;

   struct element *Next;
};

typedef struct element element;

struct matrix_node_bin
{
   int             P;
   int             Q;
   int             p;
   int             q;

   int             M;
   int             N;
   int             m;
   int             n;

   short           Row_Dis_Type;
   Cdistrib       *Row_Dis;
   void          (*Row_Distribution)();
   int           (*Row_Inverse_Distribution)();
   int           (*Row_Distribution_limit)();
   int           (*Row_Distribution_init)();
   index           Local_Max_Row;
   int             Row_Extra_param1;
   int             Row_Extra_param2;
   void           *Row_Extra;

   short           Col_Dis_Type;
   Cdistrib       *Col_Dis;
   void          (*Col_Distribution)();
   int           (*Col_Inverse_Distribution)();
   int           (*Col_Distribution_limit)();
   int           (*Col_Distribution_init)();
   index           Local_Max_Col;
   int             Col_Extra_param1;
   int             Col_Extra_param2;
   void           *Col_Extra;

   long            Pivot_Strategy;

   int             old_M;
   int             old_N;
   int             new_structure;

   ZIP_MAILER     *Node_Mailer;
   ZIP_MAILER     *Mailer;

   int             Array_len;
   int             num_element;
   int             U_elements_count;
   int             L_elements_count;
   long           *i;
   long           *j;
   double         *Value;

   CMdistrib      *Matrix_Dis;
   Dmatrix        *Csparse_Matrix;
   Smatrix        *Csparse_local_Matrix;
   DXpivot_group  *pivgrp;

   Method         *Factorization;
   Extra          *Factorization_extra;

   Method         *Solve;
   Extra          *Solve_extra;
};

typedef struct matrix_node_bin matrix_node_bin;

struct vector_node_bin
{
   int             P;
   int             Q;
   int             p;
   int             q;

   int             M;
   int             N;
   int             m;
   int             n;

   Cdistrib       *Row_Dis;
   void          (*Row_Distribution)();
   int           (*Row_Inverse_Distribution)();
   int           (*Row_Distribution_limit)();
   int           (*Row_Distribution_init)();
   void           *Row_Extra;

   Cdistrib       *Col_Dis;
   void          (*Col_Distribution)();
   int           (*Col_Inverse_Distribution)();
   int           (*Col_Distribution_limit)();
   int           (*Col_Distribution_init)();
   void           *Col_Extra;

   Cvector        *b;
   Cvector        *x;
   int             bound;
};

typedef struct vector_node_bin vector_node_bin;

#endif


void read_distribution(FILE *InputFile, short *Dis_Type,
                       char *Extra_param1_name, long *Extra_param1,
                       char *Extra_param2_name, long *Extra_param2);

void set_distribution(ZIP_MAILER *Mailer, short Dis_Type,
                      int rctype, Cdistrib **Dis, void ((**Distribution)()),
                      int ((**Inverse_Distribution)()),
                      int ((**Distribution_limit)()),
                      int ((**Distribution_init)()), int Max_Grid_dim,
                      index Grid_coor, int Global_Mat_dim, int *Local_Mat_dim,
                      int Extra_param1, int Extra_param2, void **Extra);

void Save_matrix_data_diagonal(int I, complex Diag_Value,
                               matrix_host_bin *Matrix);
void Save_matrix_data_off_diagonals(int I, int J, complex Upper_Value,
                                    complex Lower_Value,
                                    matrix_host_bin *Matrix);
void Save_matrix_data_off_diagonal(int I, int J, complex Value,
                                    matrix_host_bin *Matrix);

element *get_new_complex_element(element **Pool, int i, int j, double Real,
                                 double Imag);

void transmit_intro_letter(ZIP_MAILER *Mailer,
                           long Row_Dis_Type, long Col_Dis_Type,
                           long Row_Extra_param1, long Row_Extra_param2,
                           long Col_Extra_param1, long Col_Extra_param2,
                           long Strategy);

void receive_intro_letter(ZIP_MAILER *Mailer,
                          short *Row_Dis_Type, short *Col_Dis_Type,
                          long *Row_Extra_param1, long *Row_Extra_param2,
                          long *Col_Extra_param1, long *Col_Extra_param2,
                          long *Strategy);

void transmit_matrix_letter(ZIP_MAILER *Mailer, matrix_host_bin *Matrix);
void receive_matrix_letter(ZIP_MAILER *Mailer, matrix_node_bin *Matrix);

void transmit_timing_letter(ZIP_MAILER *Mailer, tbx_timer *Timer);
int timing_max_fn(long *ptr1, long *ptr2, int size, int nitems);
void receive_timing_letter(ZIP_MAILER *Mailer, long *sec, long *usec);

void transmit_col_vector_host(ZIP_MAILER *Mailer, vector_host_bin *Vector);
void receive_col_vector_node(ZIP_MAILER *Mailer, vector_node_bin *Vector);

void initialize_node_vector_info(matrix_node_bin *Matrix,
                                 vector_node_bin *Vector);

void add_element_real(element **Pool, element **List, int *list_count,
                      int i, int j, double Real);
void add_element_imag(element **Pool, element **List, int *list_count,
                      int i, int j, double Imag);
void Factorization(Dlu_info *Info, Cvector *b, Cvector *x, Extra *extra);
void Solve(Dlu_info *Info, Cvector *b, Cvector *x, Extra *extra);

